/* The purpose of this file is to test various hardware features.
 * A lot of effort is put into isolating these tests so that we can test all of
 * them even though one of them potentially gets stuck waiting for a message
 * that will never come.
 */

//for pthread_timedjoin_np
#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <termios.h>
#include <pthread.h>
#include <unistd.h>

#include "err.h"
#include "common.h"

#define RED   "\x1B[31m"
#define GRN   "\x1B[32m"
#define YEL   "\x1B[93m"
#define RESET "\x1B[0m"

#define UART_CMD_PATH "/dev/ttyS1"
#define UART_PRINTF_PATH "/dev/ttyS0"
#define ICOUNT_PATH "/dev/icount"
#define MAC_PATH "/sys/class/net/eth0/address"
#define BUF_PATH "/sys/class/gpio/buf_en/value"
#define DESC_MAX_SIZE 128
#define MAC_LEN_CHARS (17)
#define BUFFER_SIZE 1024
#define TASK_TIMEOUT_S (120)
#define ICOUNT_SLEEP_S 5
#define ETHERNET_SLEEP_S 30
#define USB_MOUNT_MAX_TRIES 10
#define USB_TEST_SLEEP_S 2
#define USB_TRANSMISSION_SIZE_BYTES (1024 * 1024 * 1)

#define printf_action(format, ...) printf(YEL format RESET, ##__VA_ARGS__);

#define STATUS_LED_SERVICE  "/etc/init.d/status-led"
#define SERVICE_DISABLE     "disable"
#define SERVICE_ENABLE      "enable"
#define SERVICE_STOP        "stop"
#define SERVICE_START       "start"
#define STATUS_LED_TRIGGER_FILE "/sys/class/leds/led1/trigger"
#define STATUS_LED_TRIGGER_BRIGHTNESS "/sys/class/leds/led1/brightness"

typedef unsigned long long ULL;

FILE *cmd_in = NULL, *cmd_out = NULL, *printf_in = NULL;
int printf_fd, cmd_in_fd;
char buf[BUFFER_SIZE];

typedef enum {
    TEMP_TASK,
    LED_NODE_TASK,
    XTAL_TASK,
    RADIO_TASK,
    UART_TASK,
    BUF_TASK,
    NODE_ICOUNT_TASK,
    // now start tasks not involving Node
    NODE_TASKS_END,
    ICOUNT_TASK,
    USB_TASK,
    ETHERNET_TASK,
    LED_INTERNET_TASK,
    WIFI_TASK,
} task_t;

typedef enum {
    UNKNOWN_IMG,
    HWTEST_IMG,
    RADIO_SERVER_IMG,
    NULL_IMG,
} node_image_t;

node_image_t cur_image = UNKNOWN_IMG;

void print_result(const char* test_name, int res) {
    if (res)
        printf(GRN "%s test result=%d\n" RESET, test_name, res);
    else
        printf(RED "%s test result=%d\n" RESET, test_name, res);
}

void status_led_service_disable() {
    printf("Disabling Internet connection status LED service...\n");
    THR_FAIL_IF(system(STATUS_LED_SERVICE " " SERVICE_STOP) != 0);
    THR_FAIL_IF(system("echo none > " STATUS_LED_TRIGGER_FILE) != 0); 
    THR_FAIL_IF(system("echo 0 > " STATUS_LED_TRIGGER_BRIGHTNESS) != 0); 
}

void status_led_switch_on() {
    THR_FAIL_IF(system("echo 1 > " STATUS_LED_TRIGGER_BRIGHTNESS) != 0); 
}

void status_led_switch_off() {
    THR_FAIL_IF(system("echo 0 > " STATUS_LED_TRIGGER_BRIGHTNESS) != 0); 
}

void status_led_service_restore() {
    printf("Restoring Internet connection status LED service...\n");
    THR_FAIL_IF(system(STATUS_LED_SERVICE " " SERVICE_START) != 0);
}

void node_set_img(node_image_t new_image) {
    if (new_image == cur_image)
        return;
    printf("Programming node with image %d\n", new_image);
    switch (new_image) {
        case HWTEST_IMG:
            THR_FAIL_IF(system("bootloader cherrymotehwtest.hex") != 0);
            break;
        case RADIO_SERVER_IMG:
            THR_FAIL_IF(system("bootloader cherrymotehwtestradioserver.hex") != 0);
            break;
        case NULL_IMG:
            THR_FAIL_IF(system("bootloader null.hex") != 0);
            break;
        default:
            THREAD_FATAL("Incorrect image specified\n");
    }
    cur_image = new_image;
}

void set_buf_val(int val) {
    FILE *fbuf;

    THR_FAIL_IF(!(fbuf = fopen(BUF_PATH, "w")));
    THR_FAIL_IF(fprintf(fbuf, "%d", val) < 0);
    THR_FAIL_IF(fclose(fbuf) != 0);
}

void send_cmd(char cmd) {
    THR_FAIL_IF(fprintf(cmd_out, "%c", cmd) < 0);
    THR_FAIL_IF(fflush(cmd_out) != 0);
}

void flush_printf_in() {
    // we need to discard buffered data in the stream as well as in the kernel
    if (printf_in)
        THR_FAIL_IF(fclose(printf_in) != 0);
    THR_FAIL_IF((printf_fd = open(UART_PRINTF_PATH, O_RDONLY)) == -1);
    THR_FAIL_IF(tcflush(printf_fd, TCIFLUSH) == -1);
    THR_FAIL_IF((printf_in = fdopen(printf_fd, "r")) == 0);
}

void flush_cmd_in() {
    // we need to discard buffered data in the stream as well as in the kernel
    if (cmd_in)
        THR_FAIL_IF(fclose(cmd_in) != 0);
    THR_FAIL_IF((cmd_in_fd = open(UART_CMD_PATH, O_RDONLY)) == -1);
    THR_FAIL_IF(tcflush(cmd_in_fd, TCIFLUSH) == -1);
    THR_FAIL_IF((cmd_in = fdopen(cmd_in_fd, "r")) == 0);
}

void reset_node() {
    THR_FAIL_IF(system("node-reset") != 0);
}

void init_node() {
    set_buf_val(1);
    node_set_img(HWTEST_IMG);
    reset_node();
    // without that there will be garbage written to cmd UART
    flush_cmd_in();
    THR_FAIL_IF((cmd_out = fopen(UART_CMD_PATH, "w")) == 0);
}

void task_cleanup(void *arg) {
    task_t task = (task_t)arg;

    if (task >= NODE_TASKS_END)
        return;

    printf("Node cleanup\n");
    // We shouldn't call pthread_exit here since this code does the cleanup - it
    // shouldn't freak out.
    if (cmd_out) {
        if (fclose(cmd_out) != 0)
            printf("fclose returned nonzero during cleanup\n");
        cmd_out = NULL;
    }
    if (cmd_in) {
        if (fclose(cmd_in) != 0)
            printf("fclose returned nonzero during cleanup\n");
        cmd_in = NULL;
    }
    if (printf_in) {
        if (fclose(printf_in) != 0)
            printf("fclose returned nonzero during cleanup\n");
        printf_in = NULL;
    }
}

size_t min(size_t a, size_t b) {
    return (a < b) ? a : b;
}

int test_uart() {
    int ret = 1;
    // since we're relying on fact that there will be query string written
    // we have to clean printf UART input from any garbage from before
    flush_printf_in();
    printf("UART test started\n");
    send_cmd(UART_CMD);
    THR_FAIL_IF(!fgets(buf, min(BUFFER_SIZE, strlen(QUERY) + 1), printf_in))
    ret = !strncmp(buf, QUERY, strlen(QUERY));
    print_result("UART", ret);
    return ret;
}

int test_temp() {
    int temp;
    int ret;

    printf("Temperature test started\n");
    send_cmd(TEMP_CMD);
    THR_FAIL_IF(fscanf(cmd_in, "%d", &temp) != 1);
    printf("Temperature is %d.%d\n", temp / 10, temp % 10);
    ret = (0 < temp && temp < 500);
    print_result("Temperature", ret);
    return ret;
}

int test_xtal() {
    int ret;

    printf("XTAL test started\n");
    send_cmd(XTAL_CMD);
    THR_FAIL_IF(fscanf(cmd_in, "%d", &ret) != 1);
    print_result("XTAL", ret);
    return ret;
}

int test_led_node() {
    int ret = 1;
    int tmp;

    printf("Node LED test started\n");
    send_cmd(LED_ON_CMD);
    THR_FAIL_IF(fscanf(cmd_in, "%d", &tmp) != 1);
    if (tmp != 1) {
        printf("Node is expected to return 1\n");
        ret = 0;
        goto test_led_ret;
    }
    printf_action("If ORANGE LED (NODE, MIDDLE) is on, press 1."
            " If LED is off, press 0. Then press ENTER.\n");
    THR_FAIL_IF(scanf("%d", &tmp) != 1);

    if (tmp != 1) {
        printf("ORANGE LED should be on but is off\n");
        ret = 0;
        // We don't trust the user and want to turn off the LED anyway.
    }

    send_cmd(LED_OFF_CMD);
    THR_FAIL_IF(fscanf(cmd_in, "%d", &tmp) != 1);
    if (tmp != 0) {
        ret = 0;
        printf("Node is expected to return 0 but returned %d\n", ret);
    }
test_led_ret:
    print_result("NODE_LED", ret);
    return ret;
}

int test_led_internet() {
    int ret = 1;
    int tmp;

    status_led_service_disable();
    printf("Internet LED test started\n");
    status_led_switch_on();
    printf_action("If GREEN LED (INTERNET, LEFT) is on, press 1. "
            "If LED is off, press 0. Then press ENTER.\n");
    THR_GOTO_IF(scanf("%d", &tmp) != 1, test_led_ret);

    if (tmp != 1) {
        printf("GREEN LED should be on but is off\n");
        ret = 0;
    }
    status_led_switch_off();

test_led_ret:
    status_led_service_restore();
    print_result("INTERNET_LED", ret);
    return ret;
}

// This test relies on the fact that Node is kept in reset when the buffers are
// disabled.
int test_buffers() {
    int ret = 1;
    int tmp;
    int i;

    printf("Buffer test started\n");
    for (i = 0; i < 2; i++) {
        if (i)
            init_node();
        else
            set_buf_val(0);
        send_cmd(LED_ON_CMD);
        // Waiting for Node to wake up
        if (i)
            THR_FAIL_IF(fscanf(cmd_in, "%d", &tmp) != 1);
        printf_action("If ORANGE LED is on, press 1. If LED is off, press 0. Then press ENTER.\n");
        THR_FAIL_IF(scanf("%d", &tmp) != 1);

        if (tmp != i) {
            printf("Buffers don't work as expected\n");
            ret = 0;
            // We don't trust the user and want to turn off the LED anyway.
        }

        send_cmd(LED_OFF_CMD);
    }
    print_result("Buffer", ret);
    return ret;
}

int test_radio() {
    int ret;

    printf("Radio test started\n");
    send_cmd(RADIO_CMD);
    THR_FAIL_IF(fscanf(cmd_in, "%d", &ret) != 1);
    print_result("Radio", ret);
    return ret;
}

void read_icount(ULL *ts, ULL *tc) {
    FILE *ifile;
    THR_FAIL_IF(!(ifile = fopen(ICOUNT_PATH, "r")));
    THR_FAIL_IF(fscanf(ifile, "%llu %llu", ts, tc) != 2);
    THR_FAIL_IF(fclose(ifile) != 0);
}

void read_icount_node(ULL *ts) {
    unsigned msb, lsb;
    send_cmd(ICOUNT_CMD);
    THR_FAIL_IF(fscanf(cmd_in, "%u, %u", &msb, &lsb) != 2);
    *ts = (((ULL)msb) << 32) + lsb;
}

void icount_measure_both(double *node_curr, double *sup_curr) {
    ULL node_tc[2];
    ULL sup_ts[2];
    ULL sup_tc[2];
    read_icount(&sup_ts[0], &sup_tc[0]);
    read_icount_node(&node_tc[0]);
    sleep(ICOUNT_SLEEP_S);
    read_icount(&sup_ts[1], &sup_tc[1]);
    read_icount_node(&node_tc[1]);
    *node_curr = (double)(node_tc[1] - node_tc[0]) / (ICOUNT_SLEEP_S * 1000 * 1000);
    *sup_curr = (double)(sup_tc[1] - sup_tc[0]) / (sup_ts[1] - sup_ts[0]);
}

double dabs(double a) {
    return a > 0 ? a : -a;
}

int test_node_icount() {
    double node_curr, sup_curr;
    double rel_error;
    int ret;

    printf("Node ICount test started\n");
    icount_measure_both(&node_curr, &sup_curr);
    rel_error = dabs(node_curr - sup_curr) / sup_curr;
    ret = rel_error < 0.5;
    print_result("Node ICount", ret);
    return ret;
}

double icount_measure() {
    ULL sup_ts[2];
    ULL sup_tc[2];
    read_icount(&sup_ts[0], &sup_tc[0]);
    sleep(ICOUNT_SLEEP_S);
    read_icount(&sup_ts[1], &sup_tc[1]);
    return (double)(sup_tc[1] - sup_tc[0]) / (sup_ts[1] - sup_ts[0]);
}

int test_icount() {
    int ret;
    double null_curr, radio_curr;

    printf("ICount test started\n");
    node_set_img(NULL_IMG);
    null_curr = icount_measure();
    node_set_img(RADIO_SERVER_IMG);
    radio_curr = icount_measure();
    // In my case radio app consumes 10x more energy. However, we'll not be that
    // strict in the test and we will only require it's at least 2x more.
    ret = null_curr > 0 && radio_curr > null_curr * 2;
    print_result("ICount", ret);
    return ret;
}


int test_usb() {
    FILE *tmp_file;
    int i;
    int ret = 0;
    int ready;
    int tries = 0;

    printf("USB test started\n");
    printf_action("Please insert the pendrive into the device's USB socket, write 1 and press ENTER.\n");
    THR_FAIL_IF(scanf("%d", &ready) != 1);
    if (!ready) {
        printf("Wrong input\n");
        goto exit;
    }
    // Pendrive should be mounted and formetted to ext4 after plugging-in by
    // hotplug.d. 
    // Run `mount` to check if it's done
    printf("Waiting for mounting pendrive into /mnt/pendrive\n");
    while(system("mount -t ext4 | grep /mnt/pendrive") != 0) {
        // We're counting attempts because this thread don't have joining timeout
        // because at the begining it's asking human to do something...
        if (tries == USB_MOUNT_MAX_TRIES) {
            printf("Pendrive hasn't been mounted!\n");
            goto exit;
        }
        sleep(USB_TEST_SLEEP_S);
        tries++;
    }
    printf("Done!\n");
    THR_FAIL_IF(!(tmp_file = fopen("/tmp/_usb_test.txt", "w")));

    for (i = 0; i < USB_TRANSMISSION_SIZE_BYTES / sizeof(int); i++) {
        int value = rand();
        THR_GOTO_IF(fwrite(&value, sizeof(value), 1, tmp_file) != 1, close_tmp);
    }
    THR_GOTO_IF(system("cd /tmp; sha256sum _usb_test.txt > _usb_test.txt.sha") != 0,
                close_tmp);
    THR_GOTO_IF(system("cp /tmp/_usb_test.txt /mnt/pendrive/") != 0, mnt_cleanup);
    THR_GOTO_IF(system("cd /mnt/pendrive/; sha256sum -c /tmp/_usb_test.txt.sha") != 0,
                mnt_cleanup);
    ret = 1;
mnt_cleanup:
    system("rm -f /mnt/pendrive/_usb_test.txt");
close_tmp:
    fclose(tmp_file);
    system("rm -f /tmp/_usb_test*");
exit:
    print_result("USB", ret);
    return ret;
}

int test_ethernet() {
    int ret;
    int ans;

    printf("Ethernet test started\n");
    printf_action("If GREEN LED (INTERNET, LEFT) is on or is blinking, press 1."
            "If LED is off, press 0. Then press ENTER.\n");
    THR_FAIL_IF(scanf("%d", &ans) != 1);

    if (ans != 1)
        printf("GREEN LED should be on\n");
    ret = ans;

    printf_action("Connect this device with the supplied one using Ethernet cable.\n");
    printf_action("Let's give the other device some time to notice the change: %ds\n",
            ETHERNET_SLEEP_S);
    sleep(ETHERNET_SLEEP_S);
    printf_action("If the GREEN LED of the supplied device is on or is blinking, "
            "press 1. Otherwise press 0. Then press ENTER.\n");
    THR_FAIL_IF(scanf("%d", &ans) != 1);
    ret = ret && ans;

    if (ans != 1)
        printf("Another device's GREEN LED should be on or be blinking\n");

    print_result("Ethernet", ret);

    return ret;
}

int test_wifi() {
    int ret = 0;

    printf("WiFi test started\n");
    THR_GOTO_IF(system("iwlist wlan0 scan | grep 'ESSID'") != 0, exit);
    ret = 1;

exit:
    print_result("WiFi", ret);
    return ret;
}

void* task_fn(void *arg) {
    long ret;
    task_t task = (task_t)arg;

    if (task < NODE_TASKS_END)
        init_node();
    pthread_cleanup_push(task_cleanup, arg);

    switch (task) {
        case TEMP_TASK:
            ret = test_temp();
            break;
        case LED_NODE_TASK:
            ret = test_led_node();
            break;
        case LED_INTERNET_TASK:
            ret = test_led_internet();
            break;
        case XTAL_TASK:
            ret = test_xtal();
            break;
        case RADIO_TASK:
            ret = test_radio();
            break;
        case UART_TASK:
            ret = test_uart();
            break;
        case BUF_TASK:
            ret = test_buffers();
            break;
        case NODE_ICOUNT_TASK:
            ret = test_node_icount();
            break;
        case ICOUNT_TASK:
            ret = test_icount();
            break;
        case USB_TASK:
            ret = test_usb();
            break;
        case ETHERNET_TASK:
            ret = test_ethernet();
            break;
        case WIFI_TASK:
            ret = test_wifi();
            break;
        default:
            ret = 0;
    }
    pthread_cleanup_pop(1);

    return (void*)(long)!ret;
}

void print_report(int mask, char *desc) {
    FILE *mac_f;
    char mac[MAC_LEN_CHARS + 1]; // + '\0'

    THR_FAIL_IF((mac_f = fopen(MAC_PATH, "r")) == 0);
    THR_FAIL_IF(fscanf(mac_f, "%17s", mac) != 1);
    THR_FAIL_IF(fclose(mac_f) != 0);

    printf("SUPERVISOR_MAC: %s RESULT_MASK: %d DESCRIPTION: %s\n", mac, mask, desc);
}

int main(int argc, char ** argv) {
    int ret;
    int mask = 0;
    char desc_string[DESC_MAX_SIZE] = "OK";
    pthread_t child_id;
    task_t tasks[] = {
        LED_INTERNET_TASK,
        LED_NODE_TASK,
        UART_TASK,
        TEMP_TASK,
        BUF_TASK,
        XTAL_TASK,
        RADIO_TASK,
        NODE_ICOUNT_TASK,
        ICOUNT_TASK,
        USB_TASK,
        ETHERNET_TASK,
        WIFI_TASK,
    };
    int tasks_use_timeout[] = {
        0, // LED_INTERNET_TASK
        0, // LED_NODE_TASK
        1, // UART_TASK
        1, // TEMP_TASK
        0, // BUF_TASK
        1, // XTAL_TASK
        1, // RADIO_TASK
        1, // NODE_ICOUNT_TASK
        1, // ICOUNT_TASK
        0, // USB_TASK
        0, // ETHERNET_TASK
        1, // WIFI_TASK
    };
    char **tasks_desc[] = {
        "LED Internet (Left, GREEN)", // LED_INTERNET_TASK
        "LED Node (Middle, ORANGE)", // LED_NODE_TASK
        "UART lite", // UART_TASK
        "Temperature sensor", // TEMP_TASK
        "UART buffer", // BUF_TASK
        "Node XTAL", // XTAL_TASK
        "Node radio", // RADIO_TASK
        "Node ICount", // NODE_ICOUNT_TASK
        "ICount", // ICOUNT_TASK
        "USB", // USB_TASK
        "Ethernet", // ETHERNET_TASK
        "WiFi", // WIFI_TASK
    };

    // It is possible that this script will be invoked through ssh.
    // Buffering doesn't work to good in that case.
    setvbuf(stdout, 0, _IONBF, 0);
    setvbuf(stderr, 0, _IONBF, 0);

    for (int i = 0; i < sizeof(tasks) / sizeof(*tasks); i++) {
        struct timespec timeout;
        
        int retval;
        task_t task = tasks[i];
        int use_timeout = tasks_use_timeout[i];
        SYS_FAIL_IF_ERR((ret = pthread_create(&child_id, NULL, &task_fn,
                                              (void*)task)) != 0, ret);
        SYS_FAIL_IF(clock_gettime(CLOCK_REALTIME, &timeout) == -1);
        timeout.tv_sec += TASK_TIMEOUT_S;
        ret = use_timeout ? pthread_timedjoin_np(child_id, (void**)&retval, &timeout) 
                : pthread_join(child_id, (void**)&retval);
        if (ret != 0) {
            printf(RED "Terminating thread, join ret code=%d\n" RESET, ret);
            SYS_FAIL_IF(pthread_cancel(child_id) != 0);
            // pthread_cancel just queues a cancel request, we still have to wait
            SYS_FAIL_IF(pthread_join(child_id, NULL) != 0);
            retval = 1;
        }
        mask |= ((retval) << i);
        if (retval) {
            // Exit loop on first failed test
            snprintf(desc_string, DESC_MAX_SIZE, "%s test failed with code %d", tasks_desc[i], ret);
            break;
        }
    }
    printf("Final mask=%d Description: %s\n", mask, desc_string);
    print_report(mask, desc_string);
    return 0;
}
