/*
 *  iCount power measurement kernel driver
 *
 *  As described in https://people.eecs.berkeley.edu/~prabal/pubs/papers/islped08icount.pdf
 *  Using PCF8593 via i2c as a counter.
 *
 *  Copyright (C) 2019 University of Warsaw
 *  Copyright (C) 2017 Maciej Dębski <md319428@students.mimuw.edu.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 * ---------------------------------------------------------------------------
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/i2c.h>
#include <linux/err.h>
#include <linux/fs.h>
#include <linux/time.h>
#include <linux/mutex.h>
#include <linux/workqueue.h>
#include <linux/delay.h>

#include "icount-i2c.h"
#include "pcf8593.h"

// #define DEBUG
#ifdef DEBUG
  #define debug(...) printk(KERN_DEBUG __VA_ARGS__)
#else
  #define debug(...) do {} while(0)
#endif

static int icount_majorNumber = 0;
static struct class* icount_chardev_class = NULL;
static struct device* icount_chardev_device = NULL;
static struct i2c_client* icount_i2c_client = NULL;
static struct i2c_data {
  struct timespec timestamp_base;
  struct timespec timestamp;
  struct timespec timeerror;

  unsigned long long overflows;
  unsigned long long last_counter;

  struct mutex lock;
} icount_data;

static struct workqueue_struct *icount_workqueue;
static struct delayed_work icount_periodic_work;
static int icount_periodic_shutdown = 0;

static struct file_operations icount_fops =
{
  .open = icount_i2c_open,
  .read = icount_i2c_read,
  .write = icount_i2c_write,
  .release = icount_i2c_release
};

static const unsigned short normal_i2c[] = { ICOUNT_I2C_ADDRESS,
							I2C_CLIENT_END };

static const struct i2c_device_id icount_id[] = {
	{ "icount", 0 },
	{ "icount-i2c", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, icount_id);

static struct i2c_driver icount_i2c_driver = {
	.driver = {
		.owner	= THIS_MODULE,
		.name	= ICOUNT_NAME,
	},
	.id_table	= icount_id,
	.probe		= icount_i2c_probe,
	.remove		= icount_i2c_remove,

	.address_list	= normal_i2c
};
module_i2c_driver(icount_i2c_driver);

static int icount_i2c_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
  int err;

  printk(KERN_INFO "icount: probe\n");

  mutex_init(&icount_data.lock);
  icount_i2c_client = client;

  err = icount_i2c_init();
  if (err < 0) {
    printk(KERN_ERR "icount: probe: i2c init failed\n");
    return err;
  }

  err = icount_chardev_init();
  if (err < 0) {
    printk(KERN_ERR "icount: probe: chardev init failed\n");
    return err;
  }

  err = icount_periodic_init();
  if (err < 0) {
    printk(KERN_ERR "icount: periodic init failed\n");
    icount_chardev_exit();
    return err;
  }

  return 0;
}

static int icount_periodic_init(void)
{
  int err;

  icount_workqueue = create_workqueue(ICOUNT_NAME "_workqueue");
  if(IS_ERR(icount_workqueue)) {
    return PTR_ERR(icount_workqueue);
  }
  INIT_DELAYED_WORK(&icount_periodic_work, icount_i2c_periodic);

  err = queue_delayed_work(icount_workqueue, &icount_periodic_work,
                           msecs_to_jiffies(ICOUNT_PERIODIC_DELAY_MS));
  if(err < 0) {
    destroy_workqueue(icount_workqueue);
    return err;
  }

  return 0;
}

static void icount_periodic_exit(void)
{
  icount_periodic_shutdown = 1;
  cancel_delayed_work(&icount_periodic_work);
  flush_workqueue(icount_workqueue);
  destroy_workqueue(icount_workqueue);
}

static void icount_i2c_periodic(struct work_struct *work)
{
 // Run fetch periodically to ensure no undetected overflows happen.
 debug("icount: periodic task\n");
 if(icount_periodic_shutdown == 0) {
   // If mutex is held, bail out, it's fine if someone else does the fetch.
   if(mutex_trylock(&icount_data.lock) == 1) {
     icount_i2c_fetch();
     mutex_unlock(&icount_data.lock);
   }
   queue_delayed_work(icount_workqueue, &icount_periodic_work,
                      msecs_to_jiffies(ICOUNT_PERIODIC_DELAY_MS));
 }
}

static int icount_i2c_set_register(u8 reg, u8 val)
{
  u8 buf[2];
  int cnt;
  buf[0] = reg;
  buf[1] = val;
  cnt = i2c_master_send(icount_i2c_client, buf, 2);
  if (cnt < 0) return cnt;
  if (cnt != 2) return -EIO;
  debug("icount: setting %02x to %02x succeeded\n", reg, val);
  return 0;
}

static int icount_i2c_get_register(u8 reg, u8 * buf, size_t len)
{
  int cnt;
  if (len < 1) return -EINVAL;
  buf[0] = reg;
  cnt = i2c_master_send(icount_i2c_client, buf, 1);
  if (cnt < 0) return cnt;
  if (cnt != 1) return -EIO;
  cnt = i2c_master_recv(icount_i2c_client, buf, len);
  if (cnt < 0) return cnt;
  if (cnt != len) return -EIO;

  debug("icount: getting %d from %02x succeeded: %02x %02x %02x %02x %02x %02x %02x [...]\n", len, reg, buf[0], buf[1], buf[2], buf[3], buf[4], buf[5], buf[6]);
  return 0;
}

static int icount_i2c_init(void)
{
  int err;

  do_posix_clock_monotonic_gettime(&icount_data.timestamp_base);

  err = icount_i2c_set_register(PCF8593_CONTROL, PCF8593_CONTROL_MODE_EVENT_COUNTER | PCF8593_CONTROL_ALARM_ENABLE);
  if (err < 0) return err;
  err = icount_i2c_set_register(PCF8593_D1_D0, 0);
  if (err < 0) return err;
  err = icount_i2c_set_register(PCF8593_D3_D2, 0);
  if (err < 0) return err;
  err = icount_i2c_set_register(PCF8593_D5_D4, 0);
  if (err < 0) return err;
  err = icount_i2c_set_register(PCF8593_TIMER, 0);
  if (err < 0) return err;
  err = icount_i2c_set_register(PCF8593_ALARM, PCF8593_ALARM_TIMER_MODE_1M);

  do_posix_clock_monotonic_gettime(&icount_data.timestamp);

  return err;
}

static int icount_i2c_remove(struct i2c_client *client)
{
  printk(KERN_INFO "icount: remove\n");
  icount_periodic_exit();
  icount_i2c_client = NULL;
  icount_chardev_exit();
  return 0;
}

static int icount_i2c_open(struct inode *inodep, struct file *filep)
{
  return 0;
}

static int icount_i2c_release(struct inode *inodep, struct file *filep)
{
  return 0;
}

static int icount_i2c_fetch(void)
{
  int err;
  u8 buf[16];
  struct timespec ts2;
  unsigned long long counter = 0ULL, counter2 = 0ULL;

  do_posix_clock_monotonic_gettime(&icount_data.timestamp);

  err = icount_i2c_get_register(0x01, buf, 7);
  if (err < 0) return err;

  counter = pcf8593_combine_8_digits(buf[6], buf[2], buf[1], buf[0]);

  debug("icount: fetch old: %llu new: %llu overflows: %llu\n", icount_data.last_counter, counter, icount_data.overflows);

  if(counter < icount_data.last_counter) {

    usleep_range(500, 2000);

    err = icount_i2c_get_register(0x01, buf, 7);
    if (err < 0) return err;
    counter2 = pcf8593_combine_8_digits(buf[6], buf[2], buf[1], buf[0]);

    if(counter2 < icount_data.last_counter) {
      icount_data.overflows += 1;
      printk(KERN_INFO "icount: Overflow %llu -> %llu -> %llu (ts: %llu, overflows: %llu)\n",
             icount_data.last_counter, counter, counter2, icount_i2c_format_ts(), icount_data.overflows);
    } else {
      printk(KERN_WARNING "icount: Unconfirmed overflow %llu -> %llu -> %llu (ts: %llu, overflows: %llu)\n",
             icount_data.last_counter, counter, counter2, icount_i2c_format_ts(), icount_data.overflows);
    }
    counter = counter2;
  }
  icount_data.last_counter = counter;

  do_posix_clock_monotonic_gettime(&ts2);

  icount_data.timeerror = timespec_sub(ts2, icount_data.timestamp);

  return 0;
}

static unsigned long long icount_i2c_format_ts(void)
{
  struct timespec timediff;
  timediff = timespec_sub(icount_data.timestamp, icount_data.timestamp_base);
  return ((unsigned long long) timediff.tv_sec) * USEC_PER_SEC +
         ((unsigned long long) (timediff.tv_nsec / NSEC_PER_USEC));
}

static unsigned long long icount_i2c_format_value(void)
{
  return icount_data.overflows * PCF8593_OVERFLOW + icount_data.last_counter;
}

static ssize_t icount_i2c_read(struct file *filep, char *buffer, size_t len, loff_t *offset)
{
  char icount_buffer[64];
  size_t icount_buffer_len;
  int err;

  err = mutex_lock_interruptible(&icount_data.lock);
  if (err < 0) return err;

  err = icount_i2c_fetch();
  if (err < 0) {
    mutex_unlock(&icount_data.lock);
    return err;
  }
  icount_buffer_len = snprintf(icount_buffer, 64, "%llu %llu\n", icount_i2c_format_ts(), icount_i2c_format_value());

  mutex_unlock(&icount_data.lock);

  if (len < icount_buffer_len) {
    // Partial reads are not supported.
    return -EAGAIN;
  }
  copy_to_user(buffer, icount_buffer, icount_buffer_len);
  return icount_buffer_len;
}

static ssize_t icount_i2c_write(struct file *filep, const char *buffer, size_t len, loff_t *offset)
{
  // Writes are not allowed.
  return -EINVAL;
}

static int icount_chardev_init(void)
{
  icount_majorNumber = register_chrdev(0, ICOUNT_DEVICE_NAME, &icount_fops);
  if (icount_majorNumber < 0) {
    printk(KERN_ERR "icount: failed to register a major number\n");
    return icount_majorNumber;
  }

  icount_chardev_class = class_create(THIS_MODULE, ICOUNT_CLASS_NAME);
  if (IS_ERR(icount_chardev_class)) {
    unregister_chrdev(icount_majorNumber, ICOUNT_DEVICE_NAME);

    printk(KERN_ERR "icount: failed to create class\n");
    return PTR_ERR(icount_chardev_class);
  }

  icount_chardev_device = device_create(icount_chardev_class, NULL,
      MKDEV(icount_majorNumber, 0), NULL, ICOUNT_DEVICE_NAME);
  if (IS_ERR(icount_chardev_device)) {
    class_destroy(icount_chardev_class);
    unregister_chrdev(icount_majorNumber, ICOUNT_DEVICE_NAME);

    printk(KERN_ERR "icount: failed to create device\n");
    return PTR_ERR(icount_chardev_class);
  }

  printk(KERN_INFO "icount: chardev initialized, majorNumber: %d\n", icount_majorNumber);
  return 0;
}

static void icount_chardev_exit(void)
{
  device_destroy(icount_chardev_class, MKDEV(icount_majorNumber, 0));
  class_unregister(icount_chardev_class);
  class_destroy(icount_chardev_class);
  unregister_chrdev(icount_majorNumber, ICOUNT_DEVICE_NAME);

  printk(KERN_INFO "icount: chardev cleaned up\n");
}

MODULE_LICENSE("GPL v2");
MODULE_AUTHOR("Maciej Dębski <md319428@students.mimuw.edu.pl>");
MODULE_DESCRIPTION(DRV_DESC);
MODULE_VERSION(DRV_VERSION);
