/*
 * buffer_monitor.c
 *
 *  Created on: Feb 15, 2017
 *      Author: P. Horban <p.horban@mimuw.edu.pl>
 */

#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>

#include "com.pb-c.h"

#include "linelist.h"

/** next, prev pointers, and lines pointer. */
static size_t ENTRY_OVERHEAD = 3 * sizeof(void*);

static pthread_mutex_t mutex;
static pthread_cond_t new_data_cond;

static llist_iter_t list_iter;

static size_t max_memory_use;

static void lock();
static void unlock();
static void rewind_iterator_to(line_seq_nr_t nr);
static void free_to_server_lines(ToServer *to_server);
static struct timespec get_timeout_instant(unsigned max_wait_time_sec);

int bm_init(size_t _max_memory_use) {
	int err = llist_init();
	if (err != 0)
		return err;

	llist_iter_init(&list_iter);

	err = pthread_mutex_init(&mutex, NULL);
	if (0 != err) {
		llist_destroy();
		return err;
	}

	err = pthread_cond_init(&new_data_cond, NULL);
	if (0 != err) {
		llist_destroy();
		pthread_mutex_destroy(&mutex);
		return err;
	}
	max_memory_use = _max_memory_use;
	return 0;
}

void bm_destory() {
	pthread_mutex_destroy(&mutex);
	pthread_cond_destroy(&new_data_cond);
	void llist_destroy();
}

void bm_set_max_memory_use(size_t _max_memory_use) {
	lock();
	max_memory_use = _max_memory_use;
	if (max_memory_use < 2 * LINE_SIZE_LIMIT)
		max_memory_use = 2 * LINE_SIZE_LIMIT;
	while (llist_byte_size() > max_memory_use && llist_size() > 1)
		llist_remove_oldest();
	unlock();
}

int bm_copy_lines(ToServer *to_server, line_seq_nr_t req_line_seq_nr,
		size_t max_lines_size, unsigned max_wait_time_sec) {
	lock();
	assert(max_wait_time_sec > 0);

	struct timespec to = get_timeout_instant(max_wait_time_sec);
	while (llist_is_empty() || req_line_seq_nr == llist_end_seq_nr()) {
		int err = pthread_cond_timedwait(&new_data_cond, &mutex, &to);
		if (err == ETIMEDOUT) {
			unlock();
			return errno = ENODATA;
		} else if (err != 0) {
			LOG_SE(FATAL, "thread_cond_timed wait error occured");
			exit(-1);
		}
	}

	if (req_line_seq_nr > llist_end_seq_nr()) {
		LOG(ERROR, "req_line_seq_nr > llist_end_seq_nr()");
		unlock();
		return EINVAL;
	}

	if (req_line_seq_nr < llist_begin_seq_nr())
		req_line_seq_nr = llist_begin_seq_nr();

	// At this point:
	// !llist_is_empty() &&
	// llist_begin_seq_nr() <= req_line_seq_nr < llist_end_seq_nr()

	to_server->has_max_available_line_seq_nr = 1;
	to_server->max_available_line_seq_nr = llist_end_seq_nr() - 1;

	to_server->has_min_available_line_seq_nr = 1;
	to_server->min_available_line_seq_nr = llist_begin_seq_nr();

	to_server->has_min_available_line_ts_ms = 1;
	to_server->min_available_line_ts_ms = llist_begin_ts_ms();

	rewind_iterator_to(req_line_seq_nr);
	to_server->has_first_line_seq_nr = 1;
	to_server->first_line_seq_nr = req_line_seq_nr;

	size_t n_lines = 1;
	size_t acc_size = llist_iter_entry_size(&list_iter) - ENTRY_OVERHEAD;
	while (1) {
		if (!llist_iter_has_next(&list_iter))
			break;
		llist_iter_step_forward(&list_iter);
		size_t item_size = llist_iter_entry_size(&list_iter) - ENTRY_OVERHEAD;
		if (acc_size + item_size > max_lines_size) {
			llist_iter_step_back(&list_iter);
			break;
		} else {
			acc_size += item_size;
			n_lines += 1;
			LOGV(TRACE3, "Adding line %zu with size %zu. Cum. size %zu",
					n_lines, item_size, acc_size);
		}
	}
	// list_iter is at the last line we want to send. We need to rewind it
	// back to first pos. We might as well begin the copying.

#define CH_MALLOCK(VAR, SZ) \
	if ((VAR = malloc(SZ)) == NULL) \
		goto _bm_copy_lines_nomem_cleanup;

	to_server->n_lines = n_lines;
	size_t line_arr_size = sizeof(LogLine**) * n_lines;

	CH_MALLOCK(to_server->lines, line_arr_size);
	memset(to_server->lines, 0, line_arr_size);

	ssize_t pos;
	for (pos = (ssize_t)n_lines - 1; pos >= 0; pos--) {
		CH_MALLOCK(to_server->lines[pos], sizeof(LogLine));
		LogLine *ll = to_server->lines[pos];
		log_line__init(ll);

		char *srcStr = llist_iter_get_line(&list_iter);
		size_t len = strlen(srcStr) + 1;
		CH_MALLOCK(ll->line, len);
		memcpy(ll->line, srcStr, len);
		ll->unix_ts_ms = llist_iter_get_timestamp(&list_iter);
		if (pos > 0)
			llist_iter_step_back(&list_iter);
	}

	unlock();
	return 0;

	_bm_copy_lines_nomem_cleanup: free_to_server_lines(to_server);
	unlock();
	return ENOMEM;
#undef CH_MALLOCK
}

/** Releases memory allocated by bm_copy_lines */
void bm_free_lines(ToServer *to_server) {
	lock();
	free_to_server_lines(to_server);
	unlock();
}

void bm_append_line(timestamp_t unix_ts_ms, char *line) {
	lock();
	int err;

	while (1) {
		err = llist_insert_new(unix_ts_ms, line);
		if (err == 0) { // Success. Now drop if over limit.
			while (llist_byte_size() > max_memory_use && llist_size() > 1)
				llist_remove_oldest();
			int cond_err = pthread_cond_signal(&new_data_cond);
			if (cond_err != 0) {
				perror("pthread_cond_signal returned error\n");
			}
			break;
		} else if (err == ENOMEM) {  // No memory left for new line.
			if (!llist_is_empty()) {  // Try dropping old line to make room.
				llist_remove_oldest();
			} else {  // List empty, yet allocation still fails. Dropping line.
				break;
			}
		} else {
			fprintf(stderr, "FATAL: Impossible return code from"
					" llist_insert_new: %d\n", err);
			exit(-1);
		}
	}

	unlock();
}

static void lock() {
	int err = pthread_mutex_lock(&mutex);
	if (err != 0) {
		perror("Failed to lock the mutex. This is fatal.\n");
		exit(-1);
	}
}

static void unlock() {
	int err = pthread_mutex_unlock(&mutex);
	if (err != 0) {
		perror("Failed to unlock the mutex. This is fatal.\n");
		exit(-1);
	}
}

static struct timespec get_timeout_instant(unsigned max_wait_time_sec) {
	int err;
	struct timespec ts;
	err = clock_gettime(CLOCK_REALTIME, &ts);
	if (err != 0) {
		perror("Clock read error. This is fatal\n");
		exit(-1);
	}
	ts.tv_sec += max_wait_time_sec;
	return ts;
}

static void free_to_server_lines(ToServer *to_server) {
	if (to_server->lines != NULL) {
		size_t pos;
		for (pos = 0; pos < to_server->n_lines; pos++) {
			LogLine *ll = to_server->lines[pos];
			if (ll != NULL) {
				if (ll->line != NULL) {
					free(ll->line);
					ll->line = NULL;
				}
				free(ll);
				to_server->lines[pos] = NULL;
			}
		}
		free(to_server->lines);
		to_server->lines = NULL;
	}
	to_server->n_lines = 0;
}

static void rewind_iterator_to(line_seq_nr_t nr) {
	assert(llist_begin_seq_nr() <= nr && nr < llist_end_seq_nr());

	if (!llist_iter_is_valid(&list_iter))
		llist_iter_set_at_begining(&list_iter);

	line_seq_nr_t cur = llist_iter_get_seq_nr(&list_iter);
	if (cur > nr)
		llist_iter_set_at_begining(&list_iter);
	cur = llist_iter_get_seq_nr(&list_iter);

	while (cur < nr) {
		llist_iter_step_forward(&list_iter);
		cur += 1;
	}

	assert(llist_iter_get_seq_nr(&list_iter) == nr);
}

