#define _DEFAULT_SOURCE // see inet_aton(2)

#include <arpa/inet.h>
#include <errno.h>
#include <poll.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>
#include <netinet/in.h>

#include "common.h"
#include "com.pb-c.h"

#define DBG(...) fprintf(stderr, __VA_ARGS__)


static bool initialized = false;
#define RECV_BUF_SIZE 1024

char *my_id_str;

int sock_fd = -1;
int cur_seq_nr;
int cur_session_id;
int ses_confirm_code;

struct sockaddr_in ser_addr;

static int try_create_sock() {
	if ((sock_fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) {
		DBG("socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP) failed\n");
		return -1;
	}

	if (connect(sock_fd, (struct sockaddr*) &ser_addr, sizeof(ser_addr))
			== -1) {
		close(sock_fd);
		sock_fd = -1;
		DBG("connect(sock_fd, &ser_addr, sizeof(ser_addr)) failed\n");
		return -1;
	}

	return 0;
}

int com_init(char *id_str, const char *ip_addr, uint16_t port) {
	if (initialized) {
		return EALREADY;
	}

	memset((char *) &ser_addr, 0, sizeof(ser_addr));
	ser_addr.sin_family = AF_INET;
	ser_addr.sin_port = htons(port);
	if (inet_aton(ip_addr, &ser_addr.sin_addr) == 0) {
		DBG("inet_aton(ip_addr , &si_other.sin_addr) failed\n");
		return -1;
	}

	cur_seq_nr = rand() % 1000000000;
	cur_session_id = rand() % 1000000000;
	ses_confirm_code = 0;
	my_id_str = id_str;

	initialized = true;
	return 0;
}

int com_next_msg_to_server(ToServer **msg_to_server) {
	*msg_to_server = malloc(sizeof(ToServer));
	if (*msg_to_server == NULL)
		return ENOMEM;

	to_server__init(*msg_to_server);

	return 0;
}

static int pb_send(ToServer *msg_to_server) {
	msg_to_server->node_id_str = my_id_str;

	msg_to_server->has_msg_seq_nr = 1;
	msg_to_server->msg_seq_nr = cur_seq_nr++;

	msg_to_server->has_session_id = 1;
	msg_to_server->session_id = cur_session_id;

	msg_to_server->has_ses_confirm_code = 1;
	msg_to_server->ses_confirm_code = ses_confirm_code;


	size_t msg_sz = to_server__get_packed_size(msg_to_server);
	LOGV(TRACE, "Will send protobuf packet of size %zu", msg_sz);
	void *pbuf = malloc(msg_sz);

	if (pbuf == NULL)
		return ENOMEM;

	size_t pck_size = to_server__pack(msg_to_server, pbuf);
	if (pck_size != msg_sz) {
		free(pbuf);
		return EBADMSG;
	}

	ssize_t sent_sz = send(sock_fd, pbuf, pck_size, 0);
	free(pbuf);

	if (sent_sz == -1) {
		LOGV(TRACE, "send() failed. error = %d\n", errno);
		return errno;
	}

	if ((size_t)sent_sz != pck_size) {
		LOGV(TRACE, "send() sent too few bytes. error = %d\n", errno);
		return EMSGSIZE;
	}

	return 0;
}

static long long time_in_ms() {
	struct timeval tv;
	gettimeofday(&tv, NULL);

	double time_in_mill = (double)tv.tv_sec * 1000. +
			(double)tv.tv_usec / 1000.;

	return (long long) time_in_mill;
}

int wait_reply(ToClient **msg_to_client, int recv_wait_ms) {
	struct pollfd fds[1];
	fds[0].fd = sock_fd;
	fds[0].events = POLLIN | POLLPRI; // check for normal or priority read

	void *pbuf = malloc(RECV_BUF_SIZE);

	if (pbuf == NULL) {
		DBG("malloc(RECV_BUF_SIZE) failed\n");
		return ENOMEM;
	}

	long long wait_start_ms = time_in_ms();

	while (true) {
		long long time_left_ms = recv_wait_ms;
		long long cur_time = time_in_ms();
		time_left_ms -= cur_time - wait_start_ms;

		if (time_left_ms > 0) {
			int pol_res = poll(fds, 1, (int)time_left_ms);
			LOGV(TRACE2, "poll(fds, 1, time_left_ms) ended after %lld "
					"returning %d",
				 time_in_ms() - cur_time, pol_res);

			if (pol_res == -1) {
				LOG(WARN, "poll(fds, 1, time_left_ms) failed");
				free(pbuf);
				return errno = EFAULT;
			}
			else if (pol_res == 0) {
				free(pbuf);
				return errno = ETIMEDOUT;
			}
		}

		ssize_t recv_sz = recv(sock_fd, pbuf, RECV_BUF_SIZE, MSG_DONTWAIT);
		LOGV(TRACE2, "recv returned %zd", recv_sz);

		if (recv_sz == -1) {
			if (time_left_ms <= 0) {
				free(pbuf);
				return errno;
			}
			else {
				continue;
			}
		}

		*msg_to_client = to_client__unpack(NULL, (size_t)recv_sz, pbuf);
		if (*msg_to_client == NULL) {
			LOG(TRACE2, "to_client__unpack(NULL, recv_sz, pbuf) == NULL - "
					"failed to unpack protobuf");
			continue;
		}

		ToClient *m = *msg_to_client;

		if (m->node_id_str == NULL || !m->has_msg_seq_nr ||
				// strcmp seems safe here, beacause protobuf library unpacked the string.
				m->msg_seq_nr != cur_seq_nr - 1 || 0 != strcmp(m->node_id_str, my_id_str)) {
			to_client__free_unpacked(m, NULL);
			m = *msg_to_client = NULL;
			continue;
		}

		if (m->has_ses_confirm_code)
			ses_confirm_code = m->ses_confirm_code;

		free(pbuf);
		return 0;
	}
}

static void close_sock() {
	if (sock_fd != -1) {
		close(sock_fd);
		sock_fd = -1;
	}
}

int com_communicate(ToServer *msg_to_server, ToClient **msg_to_client,
					int recv_wait_ms, int *actual_recv_wait_ms) {
	*actual_recv_wait_ms = 0;

	if (!initialized) {
		return EINVAL;
	}

	if (sock_fd == -1) {
		if(try_create_sock() != 0) {
			return ENOTCONN;
		}
	}


	if (pb_send(msg_to_server) != 0) {
		close_sock();
		return EAGAIN;
	}

	long long wait_start_ms = time_in_ms();
	int err = wait_reply(msg_to_client, recv_wait_ms);
	*actual_recv_wait_ms = (int)(time_in_ms() - wait_start_ms);
	if (err != 0) {
		close_sock();
		return EAGAIN;
	}

	return 0;
}

void com_release_to_server(ToServer **msg_to_server) {
	free(*msg_to_server);
	*msg_to_server = NULL;
}

void com_release_to_client(ToClient **msg_to_client) {
	to_client__free_unpacked(*msg_to_client, NULL);
	*msg_to_client = NULL;
}

void com_destroy() {
	close_sock();
	initialized = false;
}
