/*
 * linelist.c
 *
 *  Created on: 20-02-2017
 *      Author: P. Horban <p.horban@mimuw.edu.pl>
 */
#include "linelist.h"

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>

#include "list.h"

#include "common.h"


typedef struct list_el_struct {
	struct list_head list;

	char *line;
	timestamp_t ts;
} list_el_t;


static struct list_head queue;
static size_t overall_size;
static line_seq_nr_t begin_seq_nr;
static line_seq_nr_t end_seq_nr;

static size_t entry_size(list_el_t *entry);


int llist_init() {
	INIT_LIST_HEAD(&queue);
	overall_size = 0;
	begin_seq_nr = end_seq_nr = 0;
	return 0;
}

int llist_insert_new(timestamp_t ts, char *line) {
	list_el_t *el = malloc(sizeof(list_el_t));
	if (el == NULL)
		return ENOMEM;

	size_t buf_len = strlen(line) + 1;
	char *buf = malloc(buf_len);
	if (el == NULL) {
		free(el);
		return ENOMEM;
	}

	memcpy(buf, line, buf_len);

	el->line = buf;
	el->ts = ts;
	//todo dodaje nowy na koncu
	list_add_tail(&el->list, &queue);

	overall_size += entry_size(el);

	end_seq_nr += 1;
	return 0;
}

size_t llist_byte_size() {
	return overall_size;
}

line_seq_nr_t llist_begin_seq_nr() {
	assert(!llist_is_empty());
	return begin_seq_nr;
}

// todo to zwrocilo/zapisalo 1610850162, zamiast 1487669534578
line_seq_nr_t llist_begin_ts_ms(){
	assert(!llist_is_empty());
	list_el_t *el = list_entry(queue.next, list_el_t, list);
	return el->ts;
}

line_seq_nr_t llist_end_seq_nr(){
	assert(!llist_is_empty());
	return end_seq_nr;
}

int llist_is_empty(){
	return begin_seq_nr == end_seq_nr;
}

size_t llist_size(){
	return (size_t)(end_seq_nr - begin_seq_nr);
}

void llist_remove_oldest(){
	list_el_t *el = list_entry(queue.next, list_el_t, list);
	list_del(&el->list);
	overall_size -= entry_size(el);
	free(el->line);
	el->line = NULL;
	free(el);
	el = NULL;
	begin_seq_nr += 1;
}

void llist_destroy(){
	overall_size = 0;
	list_el_t *el, *tmp;
	list_for_each_entry_safe(el, tmp, &queue, list) {
		free(el->line);
		el->line = NULL;
		free(el);
		el = NULL;
	}
}

void llist_iter_init(llist_iter_t *iter) {
	iter->cur_el = NULL;
}

void llist_iter_set_at_begining(llist_iter_t *iter){
	assert(!llist_is_empty());
	iter->cur_el = list_entry(queue.next, list_el_t, list);
	iter->cur_seq_nr = begin_seq_nr;
}

int llist_iter_is_valid(llist_iter_t *iter){
	return iter->cur_el != NULL &&
			begin_seq_nr <= iter->cur_seq_nr && iter->cur_seq_nr < end_seq_nr;
}

int llist_iter_has_next(llist_iter_t *iter){
	return iter->cur_seq_nr + 1 < end_seq_nr;
}

void llist_iter_step_forward(llist_iter_t *iter){
	iter->cur_seq_nr += 1;
	iter->cur_el = list_entry(iter->cur_el->list.next, list_el_t, list);
}

void llist_iter_step_back(llist_iter_t *iter){
	iter->cur_seq_nr -= 1;
	iter->cur_el = list_entry(iter->cur_el->list.prev, list_el_t, list);
}

timestamp_t llist_iter_get_timestamp(llist_iter_t *iter){
	return iter->cur_el->ts;
}

char *llist_iter_get_line(llist_iter_t *iter) {
	return iter->cur_el->line;
}

line_seq_nr_t llist_iter_get_seq_nr(llist_iter_t *iter){
	return iter->cur_seq_nr;
}

size_t llist_iter_entry_size(llist_iter_t *iter){
	return entry_size(iter->cur_el);
}

static size_t entry_size(list_el_t *entry) {
	return sizeof(list_el_t) + strlen(entry->line) + 1;
}
