/*
 * linelist.h
 *
 *  Created on: Feb 14, 2017
 *      Author: P. Horban <p.horban@mimuw.edu.pl>
 */

#ifndef LINELIST_H_
#define LINELIST_H_

#include "common.h"

/**
 * Lines list keeps a list of dynamically allocated lines of characters. Each
 * line contains a unix epoch ms timestamp and char array. Moreover the data
 * structure keeps track of overall list size, including the metadata size
 * (pointers and timestamps) and remembers the line number of the first and last
 * line.
 *
 * Lines iterator helps to efficiently perform sequential transfer of lines as
 * it is only invalidated when the position that it points to gets poped of the
 * back of the list. Use iterator accessor methods to check if that was the
 * case.
 */

/**
 * Initializes the list.
 */
int llist_init();

/**
 * Copies contens of line and adds an entry containing it the end of the list.
 * Does not take ownership of line.
 * Returns 0 for success and ENOMEM if allocation failed.
 */
int llist_insert_new(timestamp_t ts, char *line);

/**
 * Returns the overall size of the list in bytes.
 */
size_t llist_byte_size();

/**
 * First available seq. nr. If begin == end, then none are available
 */
line_seq_nr_t llist_begin_seq_nr();
line_seq_nr_t llist_begin_ts_ms();

/**
 * Seq. nr. that will be assigned to next generated line. If last available
 * line has seq. nr. n, then llist_end_seq_nr() == n + 1.
 */
line_seq_nr_t llist_end_seq_nr();

int llist_is_empty();
size_t llist_size();

/**
 * Removes and deallocates the last element.
 */
void llist_remove_oldest();

/**
 * Deallocates all list memory.
 */
void llist_destroy();

typedef struct list_el_struct list_el_t;

typedef struct llist_iter {
	line_seq_nr_t cur_seq_nr;
	list_el_t *cur_el;
} llist_iter_t;


void llist_iter_init(llist_iter_t *iter);
void llist_iter_set_at_begining(llist_iter_t *iter);

/**
 * Checks if the line number of this iterator wasn't dropped yet.
 */
int llist_iter_is_valid(llist_iter_t *iter);

/**
 * Methods to move the iterator forward.
 */
int llist_iter_has_next(llist_iter_t *iter);
void llist_iter_step_forward(llist_iter_t *iter);
void llist_iter_step_back(llist_iter_t *iter);

/**
 * Methods to access the iterator contents.
 */
timestamp_t llist_iter_get_timestamp(llist_iter_t *iter);
char *llist_iter_get_line(llist_iter_t *iter);
line_seq_nr_t llist_iter_get_seq_nr(llist_iter_t *iter);
size_t llist_iter_entry_size(llist_iter_t *iter);


#endif /* LINELIST_H_ */
