/*
 * linetc.c
 *
 *  Created on: 17-02-2017
 *      Author: P. Horban <p.horban@mimuw.edu.pl>
 */
#include "linetc.h"

#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include "com.pb-c.h"

#include "common.h"
#include "clicom.h"
#include "buffer_monitor.h"

static linetc_conf_t c;

static double avg_wait_ms;
static line_seq_nr_t next_line_seq_nr;
static line_seq_nr_t max_line_seq_nr;
static size_t next_packet_size;

static int calc_recv_wait_ms();
static int validate_config(linetc_conf_t *conf);
static void update_wait_avg(int wait_time_ms, int actual_wait_time_ms, int err);
static void update_max_lines(int send_err);

int linetc_init(char *id_str, const char *ip_addr, int port,
		linetc_conf_t conf) {
	if (validate_config(&conf) != 0) {
		LOG(ERROR, "linetc_conf_t validation failed");
		return -1;
	}

	c = conf;
	int err = com_init(id_str, ip_addr, port);
	if (err != 0) {
		LOG(ERROR, "linetc_init failed due to com_init failure");
		return err;
	}
	avg_wait_ms = 10;
	next_line_seq_nr = max_line_seq_nr = 0;
	next_packet_size = c.min_packet_size;
	return 0;
}

void linetc_destory() {
	com_destroy();
}

void linetc_handle_one_packet(void) {
	ToServer *to_server;
	int err = com_next_msg_to_server(&to_server);
	if (err != 0) {
		LOG_SE(WARN, "com_next_msg_to_server failed. "
				"Waiting and retrying");
		sleep(1);
		return;
	}

	err = bm_copy_lines(to_server, next_line_seq_nr,
			next_packet_size, c.keep_alive_max_delay_sec);
	LOGV(TRACE, "Requested bm_copy_lines with next_packet_size: %zu",
			next_packet_size);

	if (err != 0) {
		com_release_to_server(&to_server);

		if (err == ENODATA) {
			LOG(INFO, "bm_copy_lines timeout - no lines to send. "
					"Retrying.");
		}
		else {
			LOG_SE(WARN, "bm_copy_lines failed. Waiting and retrying");
			sleep(1);
		}

		return;
	}

	if (to_server->has_first_line_seq_nr)
		max_line_seq_nr = max(max_line_seq_nr,
				to_server->first_line_seq_nr +
				(int64_t)to_server->n_lines);

	int actual_recv_wait_ms = 0;
	ToClient *to_client;

	int recv_wait_ms = calc_recv_wait_ms();
	err = com_communicate(to_server,
			&to_client,
			recv_wait_ms,
			&actual_recv_wait_ms);
	LOGV(DEBUG, "Communicate result %2d with timeout: %4d actual: %4d",
			err, recv_wait_ms, actual_recv_wait_ms);

	if (actual_recv_wait_ms < recv_wait_ms) {
		usleep((unsigned)(recv_wait_ms - actual_recv_wait_ms) * 1000);
	}

	update_wait_avg(recv_wait_ms, actual_recv_wait_ms, err);
	update_max_lines(err);

	bm_free_lines(to_server);
	com_release_to_server(&to_server);
	if (err != 0) {
		return;
	}

	if (to_client->has_req_line_seq_nr) {
		next_line_seq_nr = min(max_line_seq_nr,
				to_client->req_line_seq_nr);
	}

	com_release_to_client(&to_client);
}

int linetc_set_config(linetc_conf_t conf) {
	if (validate_config(&conf) != 0) {
		return -1;
	}
	c = conf;
	return 0;
}

static int validate_config(linetc_conf_t *conf) {
	if (conf->min_recv_wait_ms > conf->max_recv_wait_ms)
		return 1;
	if (conf->min_packet_size > conf->max_packet_size)
		return 1;
	double kc = conf->recv_wait_avg_keep_coef;
	if (0.0 >= kc || 1.0 <= kc)
		return 1;
	kc = conf->packet_shrink_multiplier;
	if (0.0 >= kc || 1.0 <= kc)
		return 1;
	if (conf->packet_size_increment == 0)
		return 1;
	if (conf->keep_alive_max_delay_sec <= 0)
		return 1;
	return 0;
}

static int calc_recv_wait_ms() {
	size_t t = (size_t)(avg_wait_ms * c.recv_wait_avg_componet +
			c.recv_wait_const_component);
	if (t < c.min_recv_wait_ms)
		t = c.min_recv_wait_ms;
	else if (t > c.max_recv_wait_ms)
		t = c.max_recv_wait_ms;
	return (int)t;
}

static void update_wait_avg(int wait_time_ms, int actual_wait_time_ms, int err) {
	double compt;
	if (err == 0) {
		compt = actual_wait_time_ms;
	}
	else {
		compt = wait_time_ms;
	}
	avg_wait_ms = avg_wait_ms * c.recv_wait_avg_keep_coef +
				(1.0 - c.recv_wait_avg_keep_coef) * compt;

	if (avg_wait_ms < c.min_recv_wait_ms)
		avg_wait_ms = (double)c.min_recv_wait_ms;
	else if (avg_wait_ms > c.max_recv_wait_ms)
		avg_wait_ms = (double)c.max_recv_wait_ms;
}

static void update_max_lines(int send_err) {
	if (send_err == 0)
		next_packet_size += c.packet_size_increment;
	else
		next_packet_size = (size_t)((double)next_packet_size *
				c.packet_shrink_multiplier);

	if (next_packet_size > c.max_packet_size)
		next_packet_size = c.max_packet_size;
	else if(next_packet_size < c.min_packet_size)
		next_packet_size = c.min_packet_size;
}
