/*
 * log_uploader.c
 *
 *  Created on: Feb 14, 2017
 *      Author: P. Horban <p.horban@mimuw.edu.pl>
 */

#include <signal.h>
#include <pthread.h>
#include <stdint.h>

#include "common.h"
#include "buffer_monitor.h"
#include "linetc.h"
#include "ttyread.h"
#include "config_file.h"

void *tr_thread_fun(void* _) {
	tr_main_loop();
	return NULL;
}

static int linetc_stopped;
static int reload_config = 0;

void stop_linetc_sig(int signo) {
	linetc_stopped = 1;
}

void reload_config_sig(int signo) {
	reload_config = 1;
}

int main(int argc, char *argv[]) {
	if (argc != 3 + 1) {
		LOG(FATAL, "Must pass exactly 3 arguments. The client id string and"
				" fifofile/tty to read and path to config file.");
		return -1;
	}

	// Randomize rand()
	srand((uint32_t)time(NULL));

	char *id_str = argv[1];
	char *upload_file = argv[2];
	char *config_file = argv[3];

	app_config_t app_config;
	if (read_config_file(config_file, &app_config) != 0) {
		LOG(FATAL, "Could not read config file");
		exit(-1);
	}

	if (bm_init(app_config.max_memory_use) != 0) {
		LOG(FATAL, "Could not initialize buffer_monitor");
		exit(-1);
	}

	if (tr_init(upload_file) != 0) {
		LOG(FATAL, "Could not initialize ttyread");
		bm_destory();
		exit(-1);
	}

	if (linetc_init(id_str, app_config.server_ip, app_config.server_port,
			app_config.linetc_conf) != 0) {
		LOG(FATAL, "Could not initialize linetc");
		tr_destroy();
		bm_destory();
		exit(-1);
	}

	pthread_t tr_thread, linetc_thread;
	pthread_attr_t attr; // thread attribute
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	if (pthread_create(&tr_thread, &attr, tr_thread_fun, NULL) != 0) {
		LOG(FATAL, "Could not create tr_thread");
		tr_destroy();
		bm_destory();
		linetc_destory();
		exit(-1);
	}

	linetc_stopped = 0;
    if (signal(SIGINT, stop_linetc_sig) == SIG_ERR ||
    	signal(SIGTERM, stop_linetc_sig) == SIG_ERR ||
		signal(SIGUSR1, reload_config_sig) == SIG_ERR) {
    	LOG(FATAL, "Could not set signal handlers")
		exit(-1);
    }

    while (!linetc_stopped) {
    	linetc_handle_one_packet();
    	if (reload_config) {
    		reload_config = 0;
			if (read_config_file(config_file, &app_config) != 0) {
				LOG(ERROR, "Could not read config file. No config update.");
			}
			else if(linetc_set_config(app_config.linetc_conf) == 0) {
				bm_set_max_memory_use(app_config.max_memory_use);
				LOG(INFO, "Config updated. Note that server IP and port was"
						" not change.");
			}
			else {
				LOG(ERROR, "Config validation failed. Reload aborted");
			}
    	}
    }

	tr_stop_main();
	pthread_join(tr_thread, NULL);
	tr_destroy();
	bm_destory();
	linetc_destory();

	LOG(INFO, "Clean exit");

	return 0;
}

